;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "pcache" "20260728.1657"
  "Persistent caching for Emacs."
  '((emacs "25.1"))
  :url "https://github.com/sigma/pcache"
  :commit "17d785afa4532043afa8b2dc9ae3d9733528e758"
  :revdesc "17d785afa453"
  :keywords '("extensions")
  :authors '(("Yann Hodique" . "yann.hodique@gmail.com"))
  :maintainers '(("Yann Hodique" . "yann.hodique@gmail.com")))
